from docstring_parser.google import Section, SectionType

DEFAULT_SECTIONS = [
    Section("Arguments", "param", SectionType.MULTIPLE),
    Section("Args", "param", SectionType.MULTIPLE),
    Section("Labels", "label", SectionType.MULTIPLE),
    Section("Parameters", "param", SectionType.MULTIPLE),
    Section("Params", "param", SectionType.MULTIPLE),
    Section("Raises", "raises", SectionType.MULTIPLE),
    Section("Exceptions", "raises", SectionType.MULTIPLE),
    Section("Except", "raises", SectionType.MULTIPLE),
    Section("Attributes", "attribute", SectionType.MULTIPLE),
    Section("Example", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Example 1", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Example 2", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Example 3", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Example 4", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Example 5", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Examples", "examples", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Author", "author", SectionType.SINGULAR),
    Section("Returns", "returns", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Yields", "yields", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Input Concepts", "inputs", SectionType.MULTIPLE),
    Section("Output Concepts", "outputs", SectionType.MULTIPLE),
    Section("OAuth Arguments", "param", SectionType.MULTIPLE),
    Section("OAuth Labels", "label", SectionType.MULTIPLE),
]



from dataclasses import dataclass
from typing import List, Optional

from ...api.noun_phrase import NounPhrase
from .book_authentication_descriptor import BookAuthenticationDescriptor
from .credential_descriptor import CredentialDescriptor


@dataclass
class BookCustomAuthenticationDescriptor(BookAuthenticationDescriptor):
    noun_phrase: Optional[NounPhrase]
    credentials: List[CredentialDescriptor]



from docstring_parser.common import DocstringExample as BaseDocstringExample


class DocstringExample:
    _example: BaseDocstringExample

    def __init__(self, example: BaseDocstringExample):
        self._example = example

        if example.description:
            lines = example.description.split("\n")
            description = []
            snippet = []

            for line in lines:
                stripped_line = line.strip()
                if stripped_line.startswith(">>>") or stripped_line.startswith("..."):
                    if stripped_line.startswith(">>>"):
                        snippet.append(line.replace(">>> ", "", 1).rstrip())
                    else:
                        snippet.append(line.replace("... ", "", 1).rstrip())
                else:
                    if line:
                        description.append(line)

            self._description = "\n".join(description)
            self._snippet = "\n".join(snippet)

    @property
    def description(self) -> str:
        return self._description

    @property
    def snippet(self) -> str:
        return self._snippet



from typing import List

from .any import ConceptAnyType
from .base import ConceptType
from .optional import ConceptOptionalType, contains_optionals
from .self import ConceptSelfType
from .sensitive import ConceptSensitiveType


class ConceptUnionType(ConceptType):
    def __init__(self, inners: List[ConceptType]):
        self.inners = inners

    def simplify(self) -> ConceptType:
        flattened_inners = self.flatten()
        any_type = ConceptAnyType() in flattened_inners

        if len(flattened_inners) == 1:
            simplified = flattened_inners.pop()
        elif any_type:
            simplified = ConceptAnyType()
        else:
            simplified = ConceptUnionType(flattened_inners)

        # Collapse <Self> types
        inner_unions = simplified.find(ConceptUnionType)
        if inner_unions:
            for inner_union in inner_unions:
                new_simplified_union = simplified.replace(inner_union, ConceptSelfType())
                if new_simplified_union == inner_union:
                    simplified = new_simplified_union

        if contains_optionals(self):
            return ConceptOptionalType(simplified)
        return simplified

    def flatten(self, **kwargs) -> List[ConceptType]:
        flattened_inners = []
        pack_sensitives = kwargs.get("pack_sensitives", True)

        kwargs.update(pack_sensitives=False)
        for inner in self.inners:
            for concrete_type in inner.flatten(**kwargs):
                if concrete_type not in flattened_inners:
                    flattened_inners.append(concrete_type)

        if pack_sensitives:
            flattened_and_packed = []
            sensitives_accumulator = []
            for inner in flattened_inners:
                if isinstance(inner, ConceptSensitiveType):
                    for sensitive_to_be_accumulated in inner.type.flatten(**kwargs):
                        if sensitive_to_be_accumulated not in sensitives_accumulator:
                            sensitives_accumulator.append(sensitive_to_be_accumulated)
                else:
                    flattened_and_packed.append(inner)

            if sensitives_accumulator:
                if len(sensitives_accumulator) == 1:
                    packed_sensitives = ConceptSensitiveType(sensitives_accumulator.pop())
                else:
                    packed_sensitives = ConceptSensitiveType(ConceptUnionType(sensitives_accumulator))
                flattened_and_packed = [packed_sensitives] + flattened_and_packed

            flattened_inners = flattened_and_packed

        return flattened_inners  # type: ignore

    def __repr__(self) -> str:
        return f"Union[{','.join([inner.__repr__() for inner in self.inners])}]"

    def __str__(self) -> str:
        return " or ".join([f"`{inner.__str__()}`" for inner in list(self.inners)])

    def children(self) -> List[ConceptType]:
        return self.inners

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        replaced_inners = [inner.replace(old, new) for inner in self.inners]
        return ConceptUnionType(replaced_inners)

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptUnionType):
            return set(self.inners) == set(other.inners)
        return False

    def __hash__(self) -> int:
        combined_hash = hash("Union")
        for inner in self.inners:
            combined_hash ^= hash(inner)
        return combined_hash



import inspect
from typing import Any, List, Optional, Type, Union

from kognitos.bdk.api.noun_phrase import NounPhrase
from kognitos.bdk.klang import KlangParser


def concept(*args, **kwargs):  # pylint: disable=invalid-name
    cls: Optional[Type] = None
    is_a: Optional[Union[List[str], str]] = None
    unset: Optional[Any] = None

    if len(args) == 1 and not kwargs and isinstance(args[0], type):
        cls = args[0]
        is_a = None
        unset = None
    else:
        cls = None
        is_a = kwargs.get("is_a", None)
        unset = kwargs.get("unset", None)

    def decorator(cls):
        if not inspect.isclass(cls):
            raise TypeError("The concept decorator can only be applied to classes.")

        if not cls.__doc__:
            raise ValueError("missing docstring")

        if not is_a:
            raise ValueError("missing 'is_a'")

        # parse noun phrase
        noun_phrases_is_a: List[NounPhrase] = []
        for noun_phrase in is_a if isinstance(is_a, list) else [is_a]:
            noun_phrases_tuples, _ = KlangParser.parse_noun_phrases(noun_phrase)
            if len(noun_phrases_tuples) != 1:
                raise ValueError("is_a cannot contain possessive suffix")

            noun_phrases_is_a.append(NounPhrase.from_tuple(noun_phrases_tuples[0]))

        if not hasattr(cls, "__is_a__"):
            cls.__is_a__ = set(noun_phrases_is_a)

        if not hasattr(cls, "__unset__") and unset is not None:
            cls.__unset__ = unset

        return cls

    if cls is None:
        return decorator

    return decorator(cls)



import re
from dataclasses import dataclass
from typing import List, Optional, Tuple


@dataclass
class NounPhrase:
    """Represents a noun phrase.

    Args:
        head (str): The head of the noun phrase.
        modifiers (List[str], optional): The modifiers of the noun phrase. Defaults to None.

    Methods:
        __str__(): Returns a string representation of the noun phrase.
        to_string(): Returns a string representation of the noun phrase.

    Examples:
        # Create a noun phrase with the head "cat"
        >>> np = NounPhrase("cat")
        >>> print(np)
        cat

        # Create a noun phrase with the head "dog" and modifiers ["big", "white"]
        >>> np = NounPhrase("dog", ["big", "white"])
        >>> print(np)
        big white dog

    """

    head: str
    modifiers: Optional[List[str]] = None

    def __eq__(self, other):
        if self.head != other.head:
            return False

        self_modifiers = self.modifiers if self.modifiers else []
        other_modifiers = other.modifiers if other.modifiers else []

        return self_modifiers == other_modifiers

    def __str__(self):
        return self.to_string()

    def to_string(self):
        return " ".join((self.modifiers if self.modifiers else []) + [self.head])

    @classmethod
    def from_snake_case(cls, snake_str: str) -> "NounPhrase":
        # replace underscores with spaces
        words = snake_str.replace("_", " ")
        return NounPhrase(head=words, modifiers=[])

    @classmethod
    def from_pascal_case(cls, pascal_str: str) -> "NounPhrase":
        # remove existing spaces
        pascal_str = pascal_str.replace(" ", "")
        # insert space before each capital letter (except the first one)
        words = re.sub(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])", " ", pascal_str).lower()
        return NounPhrase(head=words, modifiers=[])

    @classmethod
    def from_head(cls, head: str) -> "NounPhrase":
        return NounPhrase(head=head, modifiers=[])

    @classmethod
    def from_str(cls, string: str) -> "NounPhrase":
        return NounPhrase.from_word_list(string.split(" "))

    @classmethod
    def from_word_list(cls, word_list: List[str]) -> "NounPhrase":
        return NounPhrase(head=word_list[-1], modifiers=word_list[0:-1])

    @classmethod
    def from_tuple(cls, noun_phrase_tuple: Tuple[str, Optional[List[str]]]) -> "NounPhrase":
        return NounPhrase(noun_phrase_tuple[0], noun_phrase_tuple[1])

    def to_camel_case(self) -> str:
        snake_str = self.to_snake_case()
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    def to_snake_case(self) -> str:
        text = self.to_string()
        return text.lower().replace(" ", "_")

    def to_kebab_case(self) -> str:
        snake_str = self.to_snake_case()
        return snake_str.replace("_", "-")

    def to_field_names(self) -> List[str]:
        snake_case = self.to_snake_case()
        camel_case = self.to_camel_case()
        kebab_case = self.to_kebab_case()
        return [snake_case, camel_case, kebab_case]

    def __hash__(self):
        return hash((self.head, tuple(self.modifiers) if self.modifiers else ()))



from dataclasses import dataclass
from typing import Any, List, Optional

from ..api.noun_phrase import NounPhrase
from .types import ConceptOptionalType, ConceptType


@dataclass
class ConceptDescriptor:
    noun_phrases: List[NounPhrase]
    type: ConceptType
    default_value: Any
    description: Optional[str] = None

    @property
    def is_optional(self) -> bool:
        return isinstance(self.type, ConceptOptionalType)



class SignatureError(Exception):
    """
    Raised by BDK's type checkers when there is a mismatch between the english signature and the
    python signature of a procedure
    """

    def __init__(self, msg: str):
        super().__init__(msg)



class CredentialType:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return False


class CredentialScalarType:
    TEXT = CredentialType("Text")
    SENSITIVE = CredentialType("Sensitive")



from __future__ import annotations

import inspect
from inspect import Signature
from types import NoneType
from typing import Dict, List, Optional, Tuple, Union

from ...api import FilterExpression, NounPhrase
from ...docstring import Docstring
from ...errors import SignatureError
from ..book_procedure_descriptor import BookProcedureDescriptor
from ..book_procedure_signature import BookProcedureSignature
from ..concept_descriptor import ConceptDescriptor
from ..example import Example
from ..factory.parameter_concept import ParameterConceptFactory
from ..parameter_concept_bind import ParameterConceptBind
from .concept import ConceptFactory


def noun_phrase_to_parameter(noun_phrase: NounPhrase) -> str:
    name = noun_phrase.modifiers.copy() if noun_phrase.modifiers else []
    name.append(noun_phrase.head)

    return "_".join(name)


def find_noun_phrase_path(target_or_object: Optional[List[NounPhrase]], noun_phrases: List[NounPhrase]) -> Optional[List[NounPhrase]]:
    if not target_or_object:
        return None

    len_target_or_object = len(target_or_object)
    len_noun_phrases = len(noun_phrases)

    for start in range(len_target_or_object - len_noun_phrases + 1):
        if target_or_object[start : start + len_noun_phrases] == noun_phrases:
            return target_or_object[: start + len_noun_phrases]

    return None


def argument_name(python_signature: Signature, pos: Optional[int]) -> Optional[str]:
    """
    Get the name of the argument at the given position in the signature.

    Args:
        python_signature: The signature to get the argument name from.
        pos: The position of the argument to get the name from.

    Returns:
        The name of the argument at the given position, or None if the position is out of bounds.
    """
    if pos is None:
        return None

    parameters = list(python_signature.parameters.values())

    if len(parameters) > pos:
        return parameters[pos].name

    return None


def has_default(python_signature: Signature, pos: Optional[int]) -> Optional[bool]:
    """
    Check if the argument at the given position in the signature has a default value.

    Args:
        python_signature: The signature to check the default value of the argument from.
        pos: The position of the argument to check the default value of.

    Returns:
        True if the argument has a default value, False otherwise. If the position is out of bounds, None is returned.
    """
    if pos is None:
        return None

    parameters = list(python_signature.parameters.values())

    if 0 <= pos < len(parameters):
        return parameters[pos].default is not inspect.Parameter.empty

    return None


class BookProcedureFactory:
    @classmethod
    def create(
        cls, identifier: str, english_signature: BookProcedureSignature, python_signature: Signature, docstring: Docstring, override_connection_required: bool
    ) -> BookProcedureDescriptor:
        # by default no procedure is filter enabled
        filter_capable = False
        filter_argument_position: Optional[int] = None

        # by default no procedure is page enabled
        page_capable = False
        offset_argument_position: Optional[int] = None
        limit_argument_position: Optional[int] = None

        # parameter concept map
        pcm: List[ParameterConceptBind] = []

        # add mapping for each part of speech
        for pos in ("object", "target"):
            english_pos: Optional[List[NounPhrase]] = getattr(english_signature, pos, None)
            if english_pos:
                python_pos_parameter = python_signature.parameters.get(pos)
                if python_pos_parameter:
                    parameter_pos_annotation = getattr(
                        python_pos_parameter.annotation,
                        "__origin__",
                        python_pos_parameter.annotation,
                    )
                    if parameter_pos_annotation and issubclass(parameter_pos_annotation, Tuple):
                        concepts = [
                            ConceptFactory.from_noun_phrase_and_annotation(
                                noun_phrase, python_pos_parameter.annotation.__args__[idx], description=docstring.input_description_by_noun_phrases([noun_phrase])
                            )
                            for idx, noun_phrase in enumerate(english_pos)
                        ]

                        pcm.append(ParameterConceptFactory.from_parameter(python_pos_parameter, concepts=concepts, description=docstring.param_description_by_name(pos)))
                    else:
                        pos_noun_phrase = NounPhrase(head=pos, modifiers=[])
                        concept = ConceptFactory.from_noun_phrase_and_annotation(pos_noun_phrase, python_pos_parameter.annotation)
                        for noun_phrases in [[pos_noun_phrase], english_pos]:
                            if noun_phrases:
                                concept.description = docstring.input_description_by_noun_phrases(noun_phrases)
                                if concept.description:
                                    break

                        pcm.append(ParameterConceptFactory.from_parameter(python_pos_parameter, concepts=[concept], description=docstring.param_description_by_name(pos)))

        # add mapping for any remaining input parameter
        for idx, p in enumerate(python_signature.parameters.items()):
            _, parameter = p

            if parameter.name == "self":
                continue

            if parameter.annotation == FilterExpression or (
                hasattr(parameter.annotation, "__origin__")
                and parameter.annotation.__origin__ == Union
                and len(parameter.annotation.__args__) == 2
                and parameter.annotation.__args__[0] == FilterExpression
                and parameter.annotation.__args__[1] == NoneType
            ):
                filter_argument_position = idx
                filter_capable = True
                continue

            if parameter.annotation == int or (
                hasattr(parameter.annotation, "__origin__")
                and parameter.annotation.__origin__ == Union
                and len(parameter.annotation.__args__) == 2
                and parameter.annotation.__args__[0] == int
                and parameter.annotation.__args__[1] == NoneType
            ):
                if parameter.name == "offset":
                    offset_argument_position = idx
                    if limit_argument_position is not None:
                        page_capable = True
                    continue

                if parameter.name == "limit":
                    limit_argument_position = idx
                    if offset_argument_position is not None:
                        page_capable = True
                    continue

            if parameter.name not in [param.python_name for param in pcm]:
                parameter_concept = ParameterConceptFactory.from_parameter(parameter, description=docstring.param_description_by_name(parameter.name))
                for concept in parameter_concept.concepts:
                    for noun_phrases in [
                        find_noun_phrase_path(english_signature.object, concept.noun_phrases),
                        find_noun_phrase_path(english_signature.target, concept.noun_phrases),
                        concept.noun_phrases,
                    ]:
                        if noun_phrases:
                            concept.description = docstring.input_description_by_noun_phrases(noun_phrases)
                            if not concept.description:
                                concept.description = parameter_concept.description

                            if concept.description:
                                break
                pcm.append(parameter_concept)

        # add type for outputs
        outputs: List[ConceptDescriptor] = []
        return_annotation = python_signature.return_annotation
        if english_signature and english_signature.outputs:
            if len(english_signature.outputs) > 1 and len(english_signature.outputs) != len(return_annotation.__args__):
                raise SignatureError("the number of elements in the return tuple do not match the number of outputs in the english signature")

            for idx, output in enumerate(english_signature.outputs):
                sub_annotation = return_annotation.__args__[idx] if len(english_signature.outputs) > 1 else return_annotation
                outputs.append(
                    ConceptFactory.from_noun_phrases_and_annotation(
                        output,
                        sub_annotation,
                        description=docstring.output_description_by_noun_phrases(output),
                    )
                )
        elif return_annotation and return_annotation != inspect.Signature.empty:
            answer_noun_phrase = NounPhrase(head="answer", modifiers=[])
            outputs.append(
                ConceptFactory.from_noun_phrase_and_annotation(
                    answer_noun_phrase,
                    return_annotation,
                    description=(docstring.output_description_by_noun_phrases([answer_noun_phrase])),
                )
            )

        if len(outputs) == 1 and outputs[0].description is None and docstring.returns:
            outputs[0].description = docstring.returns

        return BookProcedureDescriptor(
            id=identifier,
            english_signature=english_signature,
            parameter_concept_map=pcm,
            outputs=outputs,
            filter_capable=filter_capable,
            filter_argument_name=argument_name(python_signature, filter_argument_position),
            filter_has_default=has_default(python_signature, filter_argument_position),
            page_capable=page_capable,
            offset_argument_name=argument_name(python_signature, offset_argument_position),
            offset_has_default=has_default(python_signature, offset_argument_position),
            limit_argument_name=argument_name(python_signature, limit_argument_position),
            limit_has_default=has_default(python_signature, limit_argument_position),
            connection_required=False,
            override_connection_required=override_connection_required,
            short_description=docstring.short_description.strip() if docstring.short_description else None,
            long_description=docstring.long_description.strip() if docstring.long_description else None,
            examples=[Example.from_docstring(docstring_example) for docstring_example in docstring.examples],
        )

    @classmethod
    def create_discovered_procedure(
        cls,
        procedure_id: str,
        english: str,
        inputs: Dict[str, ConceptDescriptor],
        outputs: List[ConceptDescriptor],
        connection_required: bool = False,
        filter_capable: bool = False,
        page_capable: bool = False,
    ):
        """
        Creates a BookProcedureDescriptor for a discover function.

        Args:
            procedure_id: The identifier for the procedure. It must be unique and is expected to contain all the information required by the developer to implement the invoke function.
            english: The english description of the procedure
            inputs: The input concepts of the procedure. The keys are the names of the parameters, and the values are the ConceptDescriptors.
            outputs: The output concepts of the procedure
            connection_required: Whether the procedure requires a connection
            filter_capable: Whether the procedure supports filtering. If true a `filter_expression` argument of type FilterExpression will be added.
            page_capable: Whether the procedure supports paging. If true an `offset` and `limit` argument will be added.
        """

        english_signature = BookProcedureSignature.from_english(english)

        parameter_concept_map = [ParameterConceptBind(python_name=python_name, concepts=[concept_descriptor]) for python_name, concept_descriptor in inputs.items()]

        return BookProcedureDescriptor(
            id=procedure_id,
            english_signature=english_signature,
            outputs=outputs,
            filter_capable=filter_capable,
            page_capable=page_capable,
            parameter_concept_map=parameter_concept_map,
            connection_required=connection_required,
            override_connection_required=False,
            short_description="",
            long_description="",
            examples=[],
            is_discovered=True,
            filter_argument_name=None if not filter_capable else "filter_expression",
            filter_has_default=None if not filter_capable else False,
            offset_argument_name=None if not page_capable else "offset",
            offset_has_default=None if not page_capable else False,
            limit_argument_name=None if not page_capable else "limit",
            limit_has_default=None if not page_capable else False,
        )



from ..typing import Sensitive
from .errors import NotFoundError, TypeMismatchError
from .filter import (FilterBinaryExpression, FilterBinaryOperator,
                     FilterExpression, FilterExpressionVisitor,
                     FilterUnaryExpression, FilterUnaryOperator,
                     NounPhrasesExpression, ValueExpression)
from .noun_phrase import NounPhrase

__all__ = [
    "NotFoundError",
    "TypeMismatchError",
    "FilterBinaryExpression",
    "FilterBinaryOperator",
    "FilterExpression",
    "FilterExpressionVisitor",
    "FilterUnaryExpression",
    "FilterUnaryOperator",
    "NounPhrase",
    "Sensitive",
]



import functools
import inspect
from types import FunctionType
from typing import Optional

from ..docstring import DocstringParser
from ..reflection import CredentialDescriptor
from ..reflection.factory import BookCustomAuthenticationFactory
from ..reflection.types import CredentialScalarType
from ..typing import Sensitive


def connect(*args, **kwargs):
    fn: Optional[FunctionType] = None
    noun_phrase: Optional[str] = None
    name: Optional[str] = None

    if len(args) == 1 and not kwargs and inspect.isfunction(args[0]):
        fn = args[0]
        noun_phrase = None
        name = None
    else:
        fn = None
        noun_phrase = kwargs.get("noun_phrase", None)
        name = kwargs.get("name", None)

    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The connect decorator can only be applied to functions.")

        short_description = None
        signature = inspect.signature(fn)
        credentials = []

        if not fn.__doc__:
            raise ValueError("missing docstring")

        parsed_docstring = DocstringParser.parse(fn.__doc__)

        if parsed_docstring.short_description:
            short_description = parsed_docstring.short_description.strip()

        for param_name, parameter in signature.parameters.items():
            if param_name == "self":
                continue

            if parameter.annotation is str:
                credential_type = CredentialScalarType.TEXT
            elif parameter.annotation is Sensitive[str]:
                credential_type = CredentialScalarType.SENSITIVE
            else:
                # NOTE: We don't support other types for now
                raise TypeError(f"parameter '{param_name}' has an unsupported type")

            if parsed_docstring:
                param_docstring = parsed_docstring.param_by_name(param_name)
            else:
                param_docstring = None

            credential = CredentialDescriptor(
                id=param_name,
                label=param_docstring.label if param_docstring is not None else None,
                type=credential_type,
                description=(param_docstring.description if param_docstring is not None else None),
            )

            credentials.append(credential)

        if not hasattr(fn, "__connect__"):
            fn.__connect__ = BookCustomAuthenticationFactory.create(id=fn.__name__, noun_phrase_str=noun_phrase, credentials=credentials, description=short_description, name=name)

        if not hasattr(fn, "__signature__"):
            fn.__signature__ = signature

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    if fn is None:
        return decorator

    return decorator(fn)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date, datetime, time
from enum import Enum
from typing import Generic, List, TypeVar

from kognitos.bdk.api.noun_phrase import NounPhrase

T = TypeVar("T")


class FilterExpressionVisitor(ABC, Generic[T]):
    @abstractmethod
    def visit_binary_expression(self, expression: FilterBinaryExpression) -> T:
        pass

    @abstractmethod
    def visit_unary_expression(self, expression: FilterUnaryExpression) -> T:
        pass

    @abstractmethod
    def visit_value(self, expression: ValueExpression) -> T:
        pass

    @abstractmethod
    def visit_noun_phrases(self, expression: NounPhrasesExpression) -> T:
        pass


class FilterExpression(ABC):
    @abstractmethod
    def accept(self, visitor: FilterExpressionVisitor[T]) -> T:
        pass


class FilterBinaryOperator(Enum):
    AND = 0
    OR = 1
    EQUALS = 2
    NOT_EQUALS = 3
    IN = 4
    HAS = 5
    LESS_THAN = 6
    GREATER_THAN = 7
    LESS_THAN_OR_EQUAL = 8
    GREATER_THAN_OR_EQUAL = 9


class FilterUnaryOperator(Enum):
    NOT = 0


class FilterBinaryExpression(FilterExpression):
    def __init__(self, left: FilterExpression, operator: FilterBinaryOperator, right: FilterExpression):
        self.operator = operator
        self.left = left
        self.right = right

    def accept(self, visitor: FilterExpressionVisitor[T]) -> T:
        return visitor.visit_binary_expression(self)


class FilterUnaryExpression(FilterExpression):
    def __init__(self, operator: FilterUnaryOperator, inner: FilterExpression):
        self.operator = operator
        self.inner = inner

    def accept(self, visitor: FilterExpressionVisitor[T]) -> T:
        return visitor.visit_unary_expression(self)


class ValueExpression(FilterExpression):
    def __init__(self, value: str | int | bool | float | date | datetime | time):
        self.value = value

    def accept(self, visitor: FilterExpressionVisitor[T]) -> T:
        return visitor.visit_value(self)


class NounPhrasesExpression(FilterExpression):
    def __init__(self, noun_phrases: List[NounPhrase]):
        self.noun_phrases = noun_phrases

    def accept(self, visitor: FilterExpressionVisitor[T]) -> T:
        return visitor.visit_noun_phrases(self)



from __future__ import annotations

import dataclasses
import inspect
from functools import partial
from typing import Callable, List, TypeVar

from kognitos.bdk.api import (FilterBinaryExpression, FilterBinaryOperator,
                              FilterExpressionVisitor, FilterUnaryExpression,
                              FilterUnaryOperator, NounPhrasesExpression,
                              ValueExpression)

T = TypeVar("T")
FilterFunction = Callable[[T], bool]


class FilterPredicates:
    @staticmethod
    def filter_in(item, left, right):
        """
        Check if the right value is in the left attribute of the item.
        """
        return right in (getattr(item, left, None) or [])

    @staticmethod
    def filter_equals(item, left: str, right):
        """
        Check if the left attribute of the item is equal to the right value.
        """
        return getattr(item, left, None) == right

    @staticmethod
    def filter_not_equals(item, left, right):
        """
        Check if the left attribute of the item is not equal to the right value.
        """
        return not FilterPredicates.filter_equals(item, left, right)

    @staticmethod
    def filter_and(item, left, right):
        """
        Check if both left and right predicates are true.
        """
        return left(item) and right(item)

    @staticmethod
    def filter_or(item, left, right):
        """
        Check if either left or right predicates are true.
        """
        return left(item) or right(item)

    @staticmethod
    def filter_has(item, left, right):
        """
        Check if the left attribute of the item has the right value.
        """
        return right in (getattr(item, left, None) or [])

    @staticmethod
    def filter_less_than(item, left, right):
        """
        Check if the left attribute of the item is less than the right value.
        """
        left_value = getattr(item, left, None)
        return left_value is not None and left_value < right

    @staticmethod
    def filter_less_than_or_equal(item, left, right):
        """
        Check if the left attribute of the item is less than or equal to the right value.
        """
        left_value = getattr(item, left, None)
        return left_value is not None and left_value <= right

    @staticmethod
    def filter_greater_than(item, left, right):
        """
        Check if the left attribute of the item is greater than the right value.
        """
        left_value = getattr(item, left, None)
        return left_value is not None and left_value > right

    @staticmethod
    def filter_greater_than_or_equal(item, left, right):
        """
        Check if the left attribute of the item is greater than or equal to the right value.
        """
        left_value = getattr(item, left, None)
        return left_value is not None and left_value >= right

    @staticmethod
    def filter_not(item, predicate):
        """
        Negate the input filter predicate.
        """
        return not predicate(item)


def is_field_or_property(cls, name: str) -> bool:
    """
    Checks if the provided name is a field or a property of the given dataclass.
    """
    # Check if the name is a field
    field_names = {field.name for field in dataclasses.fields(cls)}
    if name in field_names:
        return True

    # Check if the name is a property
    for attr_name, _ in inspect.getmembers(cls, lambda a: isinstance(a, property)):
        if attr_name == name:
            return True

    return False


class CollectionFilterExpressionVisitor(FilterExpressionVisitor):
    """
    This class is an implementation of the FilterExpressionVisitor intended to be used for
    filtering in-memory collections of objects based on the provided filter expression.

    The filterable_classes parameter is a list of classes that the filter expression can be applied to.
    These are intended to be the classes included in the collection. The main objective of declaring these classes
    is to be able to handle errors when the filter expression references an attribute that does not exist in any of the classes.
    The visitor will raise a ValueError if the filter expression references an attribute that does not exist in any of the classes.

    The visitor, when accepted by the filter expression, returns a FilterFunction, defined above.
    This function is intended to be used with the filter() function to filter the collection.
    """

    binary_operator_handlers = {
        FilterBinaryOperator.EQUALS: FilterPredicates.filter_equals,
        FilterBinaryOperator.NOT_EQUALS: FilterPredicates.filter_not_equals,
        FilterBinaryOperator.IN: FilterPredicates.filter_in,
        FilterBinaryOperator.AND: FilterPredicates.filter_and,
        FilterBinaryOperator.OR: FilterPredicates.filter_or,
        FilterBinaryOperator.HAS: FilterPredicates.filter_has,
        FilterBinaryOperator.LESS_THAN: FilterPredicates.filter_less_than,
        FilterBinaryOperator.LESS_THAN_OR_EQUAL: FilterPredicates.filter_less_than_or_equal,
        FilterBinaryOperator.GREATER_THAN: FilterPredicates.filter_greater_than,
        FilterBinaryOperator.GREATER_THAN_OR_EQUAL: FilterPredicates.filter_greater_than_or_equal,
    }

    unary_operator_handlers = {
        FilterUnaryOperator.NOT: FilterPredicates.filter_not,
    }

    def __init__(self, filterable_classes: List[type]):
        self.filterable_classes = filterable_classes

    def visit_binary_expression(self, expression: FilterBinaryExpression) -> FilterFunction:
        operator = expression.operator
        left = expression.left.accept(self)  # If NounPhrase, returns str. If BinaryExpression, returns (item) => bool
        right = expression.right.accept(self)  # If ValueExpression, returns str. If BinaryExpression, returns (item) => bool
        return partial(self.binary_operator_handlers[operator], left=left, right=right)

    def visit_unary_expression(self, expression: FilterUnaryExpression) -> FilterFunction:
        operator = expression.operator
        inner = expression.inner.accept(self)
        return partial(self.unary_operator_handlers[operator], predicate=inner)

    def visit_value(self, expression: ValueExpression):
        return expression.value

    def visit_noun_phrases(self, expression: NounPhrasesExpression) -> str:
        noun_phrase = "".join(str(np) for np in expression.noun_phrases)
        if all(not is_field_or_property(cls, noun_phrase) for cls in self.filterable_classes):
            raise ValueError(f"{noun_phrase} is not a valid attribute to filter by.")
        return noun_phrase



from typing import List

from .base import ConceptType


def contains_optionals(concept_type) -> bool:
    from .sensitive import \
        ConceptSensitiveType  # pylint: disable=cyclic-import
    from .union import ConceptUnionType  # pylint: disable=cyclic-import

    if isinstance(concept_type, ConceptOptionalType):
        return True

    if isinstance(concept_type, ConceptSensitiveType):
        return contains_optionals(concept_type.type)

    if isinstance(concept_type, ConceptUnionType):
        return any(contains_optionals(inner) for inner in concept_type.inners)

    return False


class ConceptOptionalType(ConceptType):
    def __init__(self, inner: ConceptType):
        self.type = inner

    def simplify(self) -> "ConceptOptionalType":
        simplified_inner = self.type.simplify()
        if isinstance(simplified_inner, ConceptOptionalType):
            return simplified_inner
        return ConceptOptionalType(simplified_inner)

    def children(self) -> List[ConceptType]:
        return [self.type]

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return ConceptOptionalType(self.type.replace(old, new))

    def flatten(self, **kwargs) -> List[ConceptType]:
        return self.type.flatten(**kwargs)

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptOptionalType):
            return self.type == other.type
        return False

    def __repr__(self) -> str:
        return f"Optional[{self.type.name}]"

    def __str__(self) -> str:
        return f"{self.type.name}?"

    def __hash__(self) -> int:
        combined_hash = hash("Optional")
        combined_hash ^= hash(self.type)
        return combined_hash



from inspect import Parameter
from typing import List, Optional

from ...reflection import ParameterConceptBind
from ..concept_descriptor import ConceptDescriptor
from .concept import ConceptFactory


class ParameterConceptFactory:
    @classmethod
    def from_parameter(cls, parameter: Parameter, concepts: Optional[List[ConceptDescriptor]] = None, description: Optional[str] = None) -> "ParameterConceptBind":
        return ParameterConceptBind(
            python_name=parameter.name,
            concepts=concepts if concepts else [ConceptFactory.from_parameter(parameter)],
            description=description,
        )



from typing import List

from .base import ConceptType


class ConceptSelfType(ConceptType):
    def simplify(self) -> ConceptType:
        return self

    def children(self) -> List[ConceptType]:
        return []

    def __repr__(self) -> str:
        return "Self"

    def __str__(self) -> str:
        return "self"

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return self

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptSelfType):
            return True
        return False

    def __hash__(self) -> int:
        return hash("Self")



import builtins
from functools import wraps
from logging import Logger
from typing import Generic, Set, TypeVar

T = TypeVar("T")


class Sensitive(Generic[T]):
    sensitive_data: Set[str] = set()

    def __init__(self, value: T):
        Sensitive.sensitive_data.add(str(value))
        self._value = value

    def __str__(self):
        return "Sensitive(****)"

    def __repr__(self):
        return "Sensitive(****)"

    def __getattr__(self, name):
        if name == "get_value":
            return lambda: self._value
        return getattr(self._value, name)

    def __eq__(self, other):
        if isinstance(other, Sensitive):
            return self._value == other._value
        return self._value == other

    def __hash__(self):
        return hash(self._value)

    def __add__(self, other):
        return self._value + other

    def __radd__(self, other):
        return other + self._value


unwrapped_log_method = Logger._log
unwrapped_print_function = builtins.print


def scrub_filter(data: str):
    for sensitive in sorted(list(Sensitive.sensitive_data), reverse=True):
        data = str(data).replace(sensitive, "Sensitive(****)")
    return data


@wraps(unwrapped_log_method)
def scrub_sensitive_data_from_logger(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1):
    scrubbed_msg = scrub_filter(msg)
    return unwrapped_log_method(self, level, scrubbed_msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info, stacklevel=stacklevel)


@wraps(builtins.print)
def scrub_sensitive_data_from_print(*args, sep=" ", end="\n", file=None):
    scrubbed_args = (scrub_filter(arg) for arg in args)
    return unwrapped_print_function(*scrubbed_args, sep=sep, end=end, file=file)


# NOTE: These will wrap the builtin print and the log function, so that there is no way around
#       to log any sensitive data, even if the source of the logging message does not come from
#       a Sensitive object in itself.

Logger._log = scrub_sensitive_data_from_logger
builtins.print = scrub_sensitive_data_from_print
print = scrub_sensitive_data_from_print  # pylint: disable=redefined-builtin



from typing import List, Optional

from docstring_parser import Docstring as BaseDocstring
from docstring_parser import DocstringMeta as BaseDocstringMeta
from docstring_parser import DocstringParam as BaseDocstringParam

from kognitos.bdk.klang import KlangParser

from ..api.noun_phrase import NounPhrase
from .example import DocstringExample


class DocstringConcept:
    _meta: Optional[BaseDocstringMeta]

    def __init__(self, meta: Optional[BaseDocstringMeta]):
        self._meta = meta

    @property
    def name(self) -> Optional[str]:
        if self._meta:
            return self._meta.args[1]

        return None

    @property
    def description(self) -> Optional[str]:
        if self._meta:
            return self._meta.description

        return None

    @property
    def noun_phrases(self) -> Optional[List[NounPhrase]]:
        try:
            if self._meta:
                noun_phrases, _ = KlangParser.parse_determiner_noun_phrases(self._meta.args[1])
                return [NounPhrase.from_tuple(np) for np in noun_phrases] if noun_phrases else None
        except SyntaxError:
            pass

        return None


class DocstringAttribute:
    _meta: Optional[BaseDocstringMeta]

    def __init__(self, meta: Optional[BaseDocstringMeta]):
        self._meta = meta

    @property
    def name(self) -> Optional[str]:
        if self._meta:
            return self._meta.args[1]

        return None

    @property
    def description(self) -> Optional[str]:
        if self._meta:
            return self._meta.description

        return None


class DocstringParam:
    _docstring: Optional[BaseDocstringParam]
    _meta: Optional[BaseDocstringMeta]

    def __init__(self, docstring: Optional[BaseDocstringParam], meta: Optional[BaseDocstringMeta]):
        self._docstring = docstring
        self._meta = meta

    @property
    def name(self) -> Optional[str]:
        if self._docstring:
            return self._docstring.arg_name

        if self._meta:
            return self._meta.args[1]

        return None

    @property
    def description(self) -> Optional[str]:
        if self._docstring:
            return self._docstring.description

        return None

    @property
    def label(self) -> Optional[str]:
        if self._meta:
            return self._meta.description

        return None


class Docstring:
    docstring: BaseDocstring

    def __init__(self, docstring: BaseDocstring):
        self.docstring = docstring

    @property
    def author(self) -> Optional[str]:
        for meta in self.docstring.meta:
            if "author" in meta.args:
                return meta.description
        return None

    @property
    def short_description(self) -> Optional[str]:
        return self.docstring.short_description

    @property
    def long_description(self) -> Optional[str]:
        return self.docstring.long_description

    @property
    def returns(self) -> Optional[str]:
        return self.docstring.returns.description if self.docstring.returns else None

    @property
    def input_concepts(self) -> List[DocstringConcept]:
        inputs = []

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "inputs":
                    inputs.append(DocstringConcept(meta))

        return inputs

    @property
    def examples(self) -> List[DocstringExample]:
        return [DocstringExample(example) for example in self.docstring.examples]

    @property
    def output_concepts(self) -> List[DocstringConcept]:
        outputs = []

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "outputs":
                    outputs.append(DocstringConcept(meta))

        return outputs

    @property
    def attributes(self) -> List[DocstringAttribute]:
        attributes = []

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "attribute":
                    attributes.append(DocstringAttribute(meta))

        return attributes

    @property
    def params(self) -> List[DocstringParam]:
        def convert_param(param: BaseDocstringParam):
            param_meta = None
            for meta in self.docstring.meta:
                if len(meta.args) > 1:
                    if meta.args[0] == "label" and meta.args[1] == param.arg_name:
                        param_meta = meta
                        break

            return DocstringParam(param, param_meta)

        params = list(map(convert_param, self.docstring.params))

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "label":
                    already = [param for param in params if param.name == meta.args[1]]
                    if not already:
                        params.append(DocstringParam(None, meta))

        return params

    def param_by_name(self, param: str) -> Optional[DocstringParam]:
        for doc_param in self.params:
            if doc_param.name == param:
                return doc_param

        return None

    def param_description_by_name(self, param: str) -> Optional[str]:
        for doc_param in self.params:
            if doc_param.name == param:
                return doc_param.description

        return None

    def input_concept_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[DocstringConcept]:
        for doc_concept in self.input_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept

        return None

    def input_description_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[str]:
        for doc_concept in self.input_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept.description

        return None

    def output_concept_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[DocstringConcept]:
        for doc_concept in self.output_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept

        return None

    def output_description_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[str]:
        for doc_concept in self.output_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept.description

        return None



from dataclasses import dataclass
from typing import List, Optional

from .concept_descriptor import ConceptDescriptor


@dataclass
class ParameterConceptBind:
    python_name: str
    concepts: List[ConceptDescriptor]
    description: Optional[str] = None



from .authentication import (BookAuthenticationDescriptor,
                             BookCustomAuthenticationDescriptor,
                             BookOAuthAuthenticationDescriptor,
                             CredentialDescriptor, OauthArgumentDescriptor,
                             OAuthFlow, OAuthProvider, OAuthTokenFunction)
from .book_descriptor import BookDescriptor
from .book_procedure_descriptor import BookProcedureDescriptor
from .book_procedure_signature import BookProcedureSignature
from .concept_descriptor import ConceptDescriptor
from .parameter_concept_bind import ParameterConceptBind
from .types import (ConceptAnyType, ConceptDictionaryType, ConceptListType,
                    ConceptOpaqueType, ConceptOptionalType, ConceptScalarType,
                    ConceptSelfType, ConceptSensitiveType, ConceptTableType,
                    ConceptType, ConceptUnionType, CredentialScalarType,
                    CredentialType)



# pylint: disable=invalid-name
from __future__ import annotations

from abc import ABCMeta
from typing import List

from .base import ConceptType


class ConceptScalarTypeMeta(ABCMeta):
    @property
    def CONCEPTUAL(cls) -> ConceptScalarType:
        return ConceptScalarType("Conceptual")

    @property
    def TEXT(cls) -> ConceptScalarType:
        return ConceptScalarType("Text")

    @property
    def NUMBER(cls) -> ConceptScalarType:
        return ConceptScalarType("Number")

    @property
    def BOOLEAN(cls) -> ConceptScalarType:
        return ConceptScalarType("Boolean")

    @property
    def DATETIME(cls) -> ConceptScalarType:
        return ConceptScalarType("Datetime")

    @property
    def DATE(cls) -> ConceptScalarType:
        return ConceptScalarType("Date")

    @property
    def TIME(cls) -> ConceptScalarType:
        return ConceptScalarType("Time")

    @property
    def FILE(cls) -> ConceptScalarType:
        return ConceptScalarType("File")

    @property
    def UUID(cls) -> ConceptScalarType:
        return ConceptScalarType("UUID")


class ConceptScalarType(ConceptType, metaclass=ConceptScalarTypeMeta):
    def __init__(self, name: str):
        self._name = name

    def simplify(self) -> ConceptType:
        return self

    def children(self) -> List[ConceptType]:
        return []

    def __repr__(self) -> str:
        return self._name

    def __str__(self) -> str:
        return self._name.lower()

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return self

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptScalarType):
            return self.name == other.name
        return False

    def __hash__(self) -> int:
        return hash(self.name)



import inspect
from functools import wraps
from typing import Optional

from ..reflection import BookProcedureSignature


def blueprint(*args):
    cls = Optional[type]

    # NOTE: We only care about the first argument, the rest of args and all kwargs are ignored.
    if len(args) >= 1 and inspect.isclass(args[0]):
        cls = args[0]

        def decorator(cls):
            if not hasattr(cls.__dict__, "__is_blueprint"):
                cls.__is_blueprint = True
            return cls

        return decorator(cls)

    raise ValueError("@blueprint decorator must be applied to a class.")


def blueprint_procedure(name: str):
    # This is here to verify that the english is valid
    BookProcedureSignature.from_english(name)

    def wrapper(fn):
        return wraps(fn)(fn)

    return wrapper
